/**********************************************************************
ID: statsuite.c

Copyright 2001, Bulent Yilmaz

Part of Microbrew MicroSieve
USENET SPAM filter

http://www.microbrew.org/products/usieve.html


MicroSieve is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

MicroSieve is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MicroSieve; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

**********************************************************************/

/**********************************************************
Reject Codes:

  -1: Auto-accepted message.
   0: Filtered & accepted message.
   1: Article exceeded maximum article size.
   2: Article auto-blackholed.
   3: Article has binary data for non-binary group.
   4: Article posted by a spam-bot.
   5: Article duplicates body of another article. 
***********************************************************/
#include "statsuite.h"

#if LOG_STATS

static unsigned long articles=0, autoaccept=0, toobig=0, blackhole=0, bnbg=0;
static unsigned long spambot=0, duplicate=0;
static int begintime=0;
static unsigned long long bytes=0;

int statlog(unsigned long artsize, int code) {

    articles++;
    bytes+=artsize;

    if (articles==1) begintime=(time(NULL)-1);

    switch (code) {
        case -1:   autoaccept++;
                   break;
        case 1:    toobig++;
                   break;        
        case 2:    blackhole++;
                   break;        
        case 3:    bnbg++;
                   break;
        case 4:    spambot++;
                   break;                   
        case 5:    duplicate++;
                   break;        
    }

    if((articles%1000)==0) statsoutput();
    
    return 0; 

}

int statsoutput() {

    FILE *temphandle;
    int timeused;

    temphandle=getloghandle();

    if(temphandle) {
        timeused=time(NULL) - begintime;
        
        fprintf(temphandle, "==============================================================================\n");
        fprintf(temphandle, "Stats Output:\n\n");

        fprintf(temphandle, "Articles: %ld \tKBytes: %llu \tTime (s): %d\n", articles, (bytes/1024), timeused);
        fprintf(temphandle, "Articles/Min: %u \tKBytes/Min: %lu\n\n", (unsigned int)((articles*60)/timeused), (unsigned long)((bytes/timeused)*(60.0/1024.0)) );

        fprintf(temphandle, "Auto-accepted: %ld (%.2f%%)\n", autoaccept, ((float)autoaccept/articles)*100.0 );
        fprintf(temphandle, "Rejections:\n");
        fprintf(temphandle, "\tMessage Exceeded MaxArticleSize:  \t%ld (%.2f%%)\n", toobig, ((float)toobig/articles)*100.0 );
        fprintf(temphandle, "\tMessage Automatically Blackholed: \t%ld (%.2f%%)\n", blackhole, ((float)blackhole/articles)*100.0 );
        fprintf(temphandle, "\tBinary in Non-binary Group:       \t%ld (%.2f%%)\n", bnbg, ((float)bnbg/articles)*100.0 );
        fprintf(temphandle, "\tMessage Posted by Spam-Bot:       \t%ld (%.2f%%)\n", spambot, ((float)spambot/articles)*100.0 );
        fprintf(temphandle, "\tMessage has Duplicate Body:       \t%ld (%.2f%%)\n", duplicate, ((float)duplicate/articles)*100.0 );

        fprintf(temphandle, "==============================================================================\n");
        fflush(temphandle);
    }

    return 0;

}

#endif /* LOG_STATS */