/**********************************************************************
ID: sumtable.c

Copyright 2001, Bulent Yilmaz

Part of Microbrew MicroSieve
USENET SPAM filter

http://www.microbrew.org/products/usieve.html


MicroSieve is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

MicroSieve is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MicroSieve; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

**********************************************************************/

#include "sumtable.h"

#if HISTORY_CHECK
typedef struct sumstore {
    unsigned long long sum;
    struct sumstore *next;
    struct sumstore *prev;
} SUMSTORE;

int tablecheck(unsigned long long checksum) {

    static long maxelt=1048575, lastelt=0;
    long frange=0, lrange=0, ccheck=0, rcheck=0;
    unsigned long long findsum=0;
    static SUMSTORE **sumarray=NULL;
    SUMSTORE *myelt=NULL;
    static SUMSTORE *oldest=NULL;
    static SUMSTORE *lastnewelt=NULL;
    int found=0;
    static SUMSTORE *storepool=NULL;
#if TESTOUTPUT
    FILE *loghandle;
    long int i;
#endif

#if TESTOUTPUT
    loghandle=getloghandle();
    fprintf(loghandle, "\ntablecheck: Processing new input: %llx\n",checksum);
    fflush(loghandle);
#endif

#if TESTOUTPUT
    if(checksum==-1) {
        cflog("Outputting data in sort order:");
        loghandle=getloghandle();
        for(i=0; i<lastelt; i++) {
            myelt=*(sumarray+i);
            fprintf(loghandle,"Pos %ld: %llx\n",i,myelt->sum);
            fflush(loghandle);
        }
        cflog("\nOutputting data in age order:");
	i=0;
	myelt=oldest;
	loghandle=getloghandle();
	while(myelt!=NULL) {
	    fprintf(loghandle,"Pos %ld: %llx\n",i,myelt->sum);
	    fflush(loghandle);
	    i++;
	    myelt=myelt->next;
        }
        
        return -1;
    }
#endif
    
    if(!sumarray) {
#if TESTOUTPUT
        loghandle=getloghandle();
        fprintf(loghandle,"tablecheck: allocating sum order array, %ld bytes.\n", ( maxelt * sizeof(SUMSTORE **) ));
        fflush(loghandle);
#endif
        sumarray=malloc( maxelt * sizeof(SUMSTORE **) );
    }
    
    if(!storepool) {
        storepool=malloc( maxelt * sizeof(SUMSTORE) );
#if TESTOUTPUT
        loghandle=getloghandle();
        fprintf(loghandle,"tablecheck: allocating data storage pool, %ld bytes.\n", ( maxelt * sizeof(SUMSTORE) ));
#endif        
    }

    frange=0;
    lrange=lastelt;
    found=0;
    
    while( ((lrange-frange)>0) && (!found) ) {
    	ccheck=frange+ceil((lrange-frange)/2);
#if TESTOUTPUT
        loghandle=getloghandle();    	
    	fprintf(loghandle, "tablecheck: Checking element %ld\n",ccheck);
#endif    	
    	myelt=*(sumarray+ccheck);
#if TESTOUTPUT
	loghandle=getloghandle();    	
    	fprintf(loghandle, "tablecheck: Comparing input %llx against this element %llx\n",checksum,myelt->sum);
#endif
        if((myelt->sum)<checksum) frange=ccheck+1;
        else if((myelt->sum)>checksum) lrange=ccheck;
        else found=1;
    }        
    
    if(found) {
    
        // Make rechecked sum youngest
            
        myelt=*(sumarray+ccheck);
        
        if((myelt->next)!=NULL) {
            if((myelt->prev)==NULL) {
                oldest=myelt->next;
                oldest->prev=NULL;
            } else {
                myelt->prev->next=myelt->next;
                myelt->next->prev=myelt->prev;            
            }
             
            lastnewelt->next=myelt;
            myelt->prev=lastnewelt;
            lastnewelt=myelt;
            myelt->next=NULL;
        }
        
#if TESTOUTPUT
        loghandle=getloghandle();
        fprintf(loghandle, "Body sum matched: %llx\n", checksum);
#endif
         
        return 0;
    
    } else {
#if TESTOUTPUT
        cflog("No matching data found, inserting data");
#endif
        if(lastelt) if((myelt->sum)<checksum) ccheck++;
        
        if(lastelt<maxelt) {
            myelt=(storepool+lastelt);
            lastelt++;
	    
            // Make room for data in array

            memmove( (sumarray+ccheck+1), (sumarray+ccheck), ( (lastelt-ccheck)*sizeof(SUMSTORE *) ) );
            
#if 0
            for(i=(lastelt-1); i>ccheck; i--) {
#if TESTOUTPUT
                loghandle=getloghandle();
                fprintf(loghandle,"Copying element %ld to %ld\n",i-1,i);
#endif
                *(sumarray+i)=*(sumarray+(i-1));

            }
#endif            

        } else {
#if TESTOUTPUT
            cflog("tablecheck: No room for new data, making some");
#endif

            frange=0;
            lrange=lastelt;
            findsum=oldest->sum;
            found=0;
            
            while( ((lrange-frange)>0) && (!found) ) {
            	rcheck=frange+ceil((lrange-frange)/2);
#if TESTOUTPUT
                loghandle=getloghandle();    	
            	fprintf(loghandle,"tablecheck: Checking element %ld\n",rcheck);
#endif    	
            	myelt=*(sumarray+rcheck);
#if TESTOUTPUT    	
                loghandle=getloghandle();
            	fprintf(loghandle,"tablecheck: Comparing input %llx against this element %llx\n",findsum,myelt->sum);
#endif
                if((myelt->sum)<findsum) frange=rcheck+1;
                else if((myelt->sum)>findsum) lrange=rcheck;
                else found=1;
            }
            
            if(oldest!=*(sumarray+rcheck)) {
                puts("Could not find oldest element, skipping");
                return -2;
            } else {

                if(rcheck>ccheck) {
                
                    memmove( (sumarray+ccheck+1), (sumarray+ccheck), ( (rcheck-ccheck)*sizeof(SUMSTORE *) ) );
#if 0
                    for(i=rcheck; i>ccheck; i--) {
#if TESTOUTPUT
                        loghandle=getloghandle();
                        fprintf(loghandle, "Copying element %ld to %ld\n",i-1,i);
#endif
                        *(sumarray+i)=*(sumarray+(i-1));
                    }
#endif
                } else if(rcheck<ccheck) {
                    ccheck--;
                    
                    memmove( (sumarray+rcheck), (sumarray+rcheck+1), ( (ccheck-rcheck)*sizeof(SUMSTORE *) ) );
#if 0
                    for(i=rcheck; i<ccheck; i++) {
#if TESTOUTPUT
                        loghandle=getloghandle();
                        fprintf(loghandle, "Copying element %ld to %ld\n",i+1,i);
#endif                      
                        *(sumarray+i)=*(sumarray+(i+1));
                    }
#endif                       
                   
                }
                
                myelt=oldest;
                oldest=myelt->next;
                oldest->prev=NULL;
            }
            
        }    
        
        // Now mangle data and shove it in ccheck.
	
	myelt->sum=checksum;

	if(lastelt>1) {
            lastnewelt->next=myelt;
	} else {
	    oldest=myelt;
            ccheck=0;
#if TESTOUTPUT	        
	    cflog("No data before, this is oldest data");
#endif	        
	}
        
        myelt->next=NULL;
        myelt->prev=lastnewelt;
        lastnewelt=myelt;
            
#if TESTOUTPUT
        loghandle=getloghandle();
        fprintf(loghandle, "Making final assignment in position %ld\n", ccheck);
#endif            
        *(sumarray+ccheck)=myelt;            
        
        return 1;
    }

    return 2;
}

long long calc_sum64(char *instring, long csize) {

    long long sum=0;
    long long *nextptr=NULL;
    long long nextnum;
    
    unsigned long int i;
    
    nextptr=(long long *)instring;
    
    if(nextptr) {
        for(i=0; i < (ceil(csize/8)); i++) {
        
            nextnum = (*(nextptr+i));
        
            sum=sum +  ( (nextnum << (sum & 63)) | (nextnum >> ( 64 - (sum & 63) )) );
        }
    }
#if TESTOUTPUT
    else cflog("Pointer did not cast properly.");
#endif
    
    return sum;
}

int summatch(char *instr, long slen) {
    
#if TESTOUTPUT
    FILE *loghandle;
#endif
    long long sum;
    char *alignstr;

#if __SOLARIS2__
    alignstr=memalign( 8, ((ceil(slen/8)+1)*8) );
#else
    alignstr=malloc( ((ceil(slen/8)+1)*8) );
#endif

#if 0
    loghandle=getloghandle();
    fprintf(loghandle, "instr: 0x%p, alignstr: 0x%p, slen: %ld\n",
        instr, alignstr, slen);
    fflush(loghandle);
#endif

    memcpy(alignstr, instr, slen);

#if 0
    loghandle=getloghandle();
    fprintf(loghandle, "instr: 0x%p, alignstr: 0x%p, slen: %ld\n",
        instr, alignstr, slen);
    fflush(loghandle);
    cflog("Past strncpy");
#endif

#if TESTOUTPUT
    cflog("summatch: About to determine article body size");
#endif
    
#if TESTOUTPUT
    loghandle=getloghandle();
    fprintf(loghandle, "summatch: Processing article body.  Size: %ld.\n", slen);
    fflush(loghandle);
#endif
    
    sum=calc_sum64(alignstr, slen);

#if TESTOUTPUT
    loghandle=getloghandle();
    fprintf(loghandle, "summatch: Sum value: %llx.\n", sum);
    fflush(loghandle);
#endif

    free(alignstr);
    
    if(sum) {
        return tablecheck(sum);
    } else return 1;
    
}
#endif /* HISTORY_CHECK */